// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit map;

interface

uses
  Classes,{$IFDEF WIN32}windows,{$ENDIF}orbtypes,map_int,SysUtils;

{$IFNDEF USELIB}

type

  TLargeInt = LARGE_INTEGER;
  Comp = LONGLONG;
  PCompList = ^TCompList;
  TCompList = array[0..MaxListSize - 1] of LONGLONG;

  TMap = class(TInterfacedObject)
  private
    FList: PCompList;
    FCount: Integer;
    FCapacity : Integer;
  protected
    procedure Grow; virtual;
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
    function Find(Start,Last,Key : DWORD): integer;
    function FindA(Start,Last,Key : DWORD): integer;
    procedure Insert(Index: Integer; Item: Comp);
    procedure _Put(Key: DWORD;Item: DWORD);
    procedure Delete(Index: Integer); virtual;
    function Expand: TMap;
    procedure clear(); virtual;
    function _Get(Key: DWORD): integer;
    function IndexOf(Item: DWORD): integer;
    procedure Pack;
    function count(): integer;
    property List: PCompList read FList;
    class procedure Error(const Msg: string; Data: Integer); virtual;
    function Remove(Key: DWORD): Integer;
  public
    destructor Destroy; override;
  end;

  TLongInterfaceMap = class(TMap,ILongInterfaceMap)
  protected
    procedure Put(Key: _ulong;Item: IUnknown);
    function Get(Key: _ulong): IUnknown;
    procedure clear(); override;
    procedure erase(Key: _ulong);
    procedure Delete(Index: Integer); override;
    function Items(Index: integer): IUnknown;
    function IndexOf(Item: IUnknown): integer;
  public
    property Item[Key: _ulong]: IUnknown read Get write Put; default;
  end;


  TInterfaceLongMap = class(TMap,IInterfaceLongMap)
  protected
    procedure Put(Key: IUnknown;Item: _ulong);
    function Get(Key: IUnknown): _ulong;
    procedure clear(); override;
    procedure Delete(Index: Integer); override;
    function IndexOf(Key: IUnknown): integer;
  public
    property Item[Key: IUnknown]: _ulong read Get write Put; default;
  end;

  TInterfaceInterfaceMap = class(TMap,IInterfaceInterfaceMap)
  protected
    procedure Put(Key: IUnknown;Item: IUnknown);
    function Get(Key: IUnknown): IUnknown;
    function IndexOf(Key: IUnknown): integer;
    property Item[Key: IUnknown]: IUnknown read Get write Put; default;
    procedure clear(); override;
    procedure Delete(Index: Integer); override;
    procedure erase(Key: IUnknown);
    function Second(Index: Integer): IUnknown;
  end;

  function long_interface_map: ILongInterfaceMap;
  function interface_long_map: IInterfaceLongMap;
  function interface_interface_map: IInterfaceInterfaceMap;

const
  SListIndexError = 'Invalid index';
  SListCapacityError = 'Capacity error';
  SListCountError = 'List error';
  REASONABLE_INT = 7;

implementation

//*********************************************************************
//   TMap
//*********************************************************************
destructor TMap.Destroy;
begin
  Clear;
end;

function TMap.count(): integer;
begin
  result := FCount;
end;

procedure TMap._Put(Key: DWORD;Item: DWORD);
var
  x,y : TLargeInt;
  ind : integer;
begin
  x.HighPart := Key;
  x.LowPart := Item;
  if FCount > 0 then
    begin
      ind := Find(0,FCount-1,Key);
      if (ind < FCount) then
        begin
          y.QuadPart := FList^[ind];
          if (DWORD(y.HighPart) = Key) then
            begin
              FList^[ind] := x.QuadPart;
              Exit;
            end;
        end;
    end
  else
    ind := 0;
  Insert(ind,x.QuadPart);
end;

procedure TMap.Clear;
begin
  SetCount(0);
  SetCapacity(0);
end;

procedure TMap.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(Comp));
end;

class procedure TMap.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

function TMap.Expand: TMap;
begin
  if FCount = FCapacity then Grow;
  Result := Self;
end;

procedure TMap.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TMap.Find(Start,Last,Key : DWORD): integer;
begin
  Result := Last;
  while True do
    begin
      Result := Last;
      if DWORD(FList^[Last] shr 32)  = Key then
        Break
      else if DWORD(FList^[Last] shr 32) < Key then
        begin
          Inc(Result);
          Break;
        end
      else if DWORD(FList^[Start] shr 32) >= Key then
        begin
          Result := Start;
          Break;
        end
      else if Start = Last then
        Break
      else
        begin
          Result := (Start+Last) div 2;
          if DWORD(FList^[Result] shr 32) <= Key then
            begin
              Dec(Last);
              Start := Result;
            end
          else
            begin
              Inc(Start);
              Last := Result - 1;
            end;
        end
    end;
end;

function TMap.Finda(Start,Last,Key : DWORD): integer;
asm
  push ebx
  push edi
  push esi
  mov esi,Start
  mov edi,Last
  mov ebx, Key
  mov ecx, [eax+4h] // FList
@@4:
  mov eax, [ecx+edi*8h+4h]
  cmp eax ,ebx
  je  @@1
  jl  @@2
  mov edx, [ecx+esi*8h+4h]
  cmp edx ,ebx
  jge @@3
  cmp esi,edi
  je @@1
  mov eax,esi
  add eax,edi
  sar eax,1
  mov edx, eax
  mov eax, [ecx+eax*8h+4h]
  cmp eax ,ebx
  jg  @@5
  dec edi
  mov esi, edx
  jmp @@4
@@5:
  inc esi
  dec edx
  mov edi,edx
  jmp @@4
@@3:
  mov edi,esi
  jmp @@1
@@2:
  inc edi
@@1:
  mov eax,edi
  pop esi
  pop edi
  pop ebx
end;

function TMap._Get(Key: DWORD): integer;
begin
  if FCount > 0 then
    begin
      Result := Find(0,FCount-1,Key);
      if (Result >= FCount) or (DWORD(FList^[Result] shr 32) <> Key) then
        Result := -1;
    end
  else
    Result := -1;
end;

function TMap.IndexOf(Item: DWORD): Integer;
begin
  Result := 0;
  while (Result < FCount) and ((FList^[Result] and $FFFFFFFF) <> Item) do
    Inc(Result);
  if Result = FCount then
    Result := -1;
end;

procedure TMap.Insert(Index: Integer; Item: Comp);
begin
  if (Index < 0) or (Index > FCount) then Error(SListIndexError, Index);
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(Comp));
  FList^[Index] := Item;
  Inc(FCount);
end;

function TMap.Remove(Key: DWORD): Integer;
begin
  Result := Find(0,FCount-1,Key);
  if (Result <> -1) and (DWORD(FList^[Result] shr 32) = Key) then
    Delete(Result)
  else
    Result := -1;
end;

procedure TMap.Pack;
var
  I: Integer;
begin
  for I := FCount - 1 downto 0 do if (FList^[I] and $FFFFFFFF) = 0 then Delete(I);
end;

procedure TMap.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > MaxListSize) then
    Error(SListCapacityError, NewCapacity);
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Comp));
    FCapacity := NewCapacity;
  end;
end;

procedure TMap.SetCount(NewCount: Integer);
begin
  if (NewCount < 0) or (NewCount > MaxListSize) then
    Error(SListCountError, NewCount);
  if NewCount > FCapacity then SetCapacity(NewCount);
  if NewCount > FCount then
    FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(Comp), 0);
  FCount := NewCount;
end;

//*********************************************************************
//  TLongInterfaceMap
//*********************************************************************
function TLongInterfaceMap.Get(Key: _ulong): IUnknown;
var
  ind : integer;
begin
  result := nil;
  ind := _Get(key);
  if ind >= 0 then
    result := IUnknown(Pointer(FList^[ind] and $FFFFFFFF));
end;

function TLongInterfaceMap.IndexOf(Item: IUnknown): integer;
begin
  result := inherited IndexOf(DWORD(Pointer(Item)));
end;

procedure TLongInterfaceMap.Put(Key: _ulong; Item: IUnknown);
var
  x,y : TLargeInt;
  ind : integer;
begin
  x.HighPart := Key;
  x.LowPart := DWORD(Pointer(Item));
  if FCount > 0 then
    begin
      ind := Find(0,FCount-1,Key);
      if (ind < FCount) then
        begin
          y.QuadPart := FList^[ind];
          if (DWORD(y.HighPart) = Key) then
            begin
              if y.LowPart <> 0 then
                IUnknown(Pointer(y.LowPart))._Release;
              if Item <> nil then
                Item._AddRef;
              FList^[ind] := x.QuadPart;
              Exit;
            end;
        end;
    end
  else
    ind := 0;
  Insert(ind,x.QuadPart);
  Item._AddRef;
end;

procedure TLongInterfaceMap.clear();
var
  i : integer;
begin
  for i := 0 to Pred(FCount) do
    if (FList^[i] and $FFFFFFFF) <> 0 then
      IUnknown(Pointer((FList^[i] and $FFFFFFFF)))._Release;
  inherited clear();
end;

procedure TLongInterfaceMap.erase(Key: _ulong);
begin
  self.Remove(Key);
end;

procedure TLongInterfaceMap.Delete(Index: Integer);
begin
  if (FList^[Index] and $FFFFFFFF) <> 0 then
    IUnknown(Pointer((FList^[Index] and $FFFFFFFF)))._Release;
  inherited Delete(Index);
end;

function TLongInterfaceMap.Items(Index: integer): IUnknown;
begin
  result := nil;
  if Index < FCount then
    result := IUnknown(Pointer((FList^[Index] and $FFFFFFFF)))
  else
    Error(SListIndexError, Index);
end;

//***************************************************************
// TInterfaceKeyMap
//***************************************************************
procedure TInterfaceLongMap.clear;
var
  i : integer;
begin
  for i := 0 to Pred(FCount) do
    if (FList^[i] shr 32) <> 0 then
      IUnknown(Pointer((FList^[i] shr 32)))._Release;
  inherited clear();
end;

procedure TInterfaceLongMap.Delete(Index: Integer);
begin
  if (FList^[Index] shr 32) <> 0 then
    IUnknown(Pointer((FList^[Index] shr 32)))._Release;
  inherited Delete(Index);
end;

function TInterfaceLongMap.Get(Key: IUnknown): _ulong;
var
  ind : integer;
begin
  result := 0;
  ind := _Get(DWORD(Pointer(key)));
  if ind >= 0 then
    result := FList^[ind] and $FFFFFFFF;
end;

function TInterfaceLongMap.IndexOf(Key: IUnknown): integer;
begin
  result := _Get(DWORD(Pointer(key)));
end;

procedure TInterfaceLongMap.Put(Key: IUnknown; Item: _ulong);
var
  x,y : TLargeInt;
  ind : integer;
begin
  x.HighPart := DWORD(Pointer(Key));
  x.LowPart := Item;
  if FCount > 0 then
    begin
      ind := Find(0,FCount-1,DWORD(Pointer(Key)));
      if (ind < FCount) then
        begin
          y.QuadPart := FList^[ind];
          if (DWORD(y.HighPart) = DWORD(Pointer(Key))) then
            begin
              FList^[ind] := x.QuadPart;
              Exit;
            end;
        end;
    end
  else
    ind := 0;
  Insert(ind,x.QuadPart);
  Key._AddRef;
end;

{ TInterfaceInterfaceMap }

function TInterfaceInterfaceMap.Get(Key: IUnknown): IUnknown;
var
  ind : integer;
begin
  result := nil;
  ind := _Get(DWORD(Pointer(key)));
  if ind >= 0 then
    result := IUnknown(Pointer(FList^[ind] and $FFFFFFFF));
end;

function TInterfaceInterfaceMap.IndexOf(Key: IUnknown): integer;
begin
  result := _Get(DWORD(Pointer(key)));
end;

procedure TInterfaceInterfaceMap.erase(Key: IUnknown);
var
  i : integer;
begin
  i := IndexOf(key);
  if i >= 0 then Delete(i)
end;

procedure TInterfaceInterfaceMap.Put(Key, Item: IUnknown);
var
  x,y : TLargeInt;
  ind : integer;
  tmp : IUnknown;
begin
  x.HighPart := DWORD(Pointer(Key));
  x.LowPart := DWORD(Pointer(Item));
  if FCount > 0 then
    begin
      ind := Find(0,FCount-1,DWORD(Pointer(Key)));
      if (ind < FCount) then
        begin
          y.QuadPart := FList^[ind];
          if (DWORD(y.HighPart) = DWORD(Pointer(Key))) then
            begin
              tmp := IUnknown(Pointer(FList^[ind] and $FFFFFFFF));
              if tmp <> nil then
                tmp._Release;
              if item <> nil then Item._AddRef;
              FList^[ind] := x.QuadPart;
              Exit;
            end;
        end;
    end
  else
    ind := 0;
  Insert(ind,x.QuadPart);
  Key._AddRef;
  if Item <> nil then
    Item._AddRef;
end;

procedure TInterfaceInterfaceMap.clear();
var
  i : integer;
begin
  for i := 0 to Pred(FCount) do
    begin
      if (FList^[i] shr 32) <> 0 then
        IUnknown(Pointer((FList^[i] shr 32)))._Release;
      if (FList^[i] and $FFFFFFFF) <> 0 then
        IUnknown(Pointer(FList^[i] and $FFFFFFFF))._Release;
    end;
  inherited clear();
end;

procedure TInterfaceInterfaceMap.Delete(Index: Integer);
begin
  if (FList^[Index] shr 32) <> 0 then
    IUnknown(Pointer((FList^[Index] shr 32)))._Release;
  if (FList^[Index] and $FFFFFFFF) <> 0 then
    IUnknown(Pointer(FList^[Index] and $FFFFFFFF))._Release;
  inherited Delete(Index);
end;

function TInterfaceInterfaceMap.Second(Index: Integer): IUnknown;
begin
  result := IUnknown(Pointer(FList^[Index] and $FFFFFFFF))
end;

function long_interface_map: ILongInterfaceMap;
begin
  result := TLongInterfaceMap.Create;
end;

function interface_long_map: IInterfaceLongMap;
begin
  result := TInterfaceLongMap.Create;
end;

function interface_interface_map: IInterfaceInterfaceMap;
begin
  result := TInterfaceInterfaceMap.Create;
end;

{$IFDEF LIBRARY}
exports
  long_interface_map,
  interface_long_map,
  interface_interface_map;
{$ENDIF}
{$ELSE}

  function long_interface_map: ILongInterfaceMap; external MTDORB_Library_Name;
  function interface_long_map: IInterfaceLongMap; external MTDORB_Library_Name;
  function interface_interface_map: IInterfaceInterfaceMap; external MTDORB_Library_Name;
implementation
{$ENDIF}
end.
